@@ -4,42 +4,22 @@ If you want to create your own data layer
 you must inherit from this base class
 """
 
-import types
-from typing import List, Optional, Tuple, Type
+from typing import Dict, List, Optional, Tuple, Type
+
+from fastapi import Request
 
 from fastapi_jsonapi.data_typing import TypeModel, TypeSchema
 from fastapi_jsonapi.querystring import QueryStringManager
 from fastapi_jsonapi.schema import BaseJSONAPIItemInSchema
+from fastapi_jsonapi.schema_builder import FieldConfig, TransferSaveWrapper
 
 
 class BaseDataLayer:
     """Base class of a data layer"""
 
-    REWRITABLE_METHODS = (
-        "query",
-        "before_create_object",
-        "after_create_object",
-        "before_get_object",
-        "after_get_object",
-        "before_get_collection",
-        "after_get_collection",
-        "before_update_object",
-        "after_update_object",
-        "before_delete_object",
-        "after_delete_object",
-        "before_create_relationship",
-        "after_create_relationship",
-        "before_get_relationship",
-        "after_get_relationship",
-        "before_update_relationship",
-        "after_update_relationship",
-        "before_delete_relationship",
-        "after_delete_relationship",
-        "retrieve_object_query",
-    )
-
     def __init__(
         self,
+        request: Request,
         schema: Type[TypeSchema],
         model: Type[TypeModel],
         url_id_field: str,
@@ -50,7 +30,7 @@ class BaseDataLayer:
         **kwargs,
     ):
         """
-
+        :param request:
         :param schema:
         :param model:
         :param url_id_field:
@@ -60,8 +40,9 @@ class BaseDataLayer:
         :param type_: resource type
         :param kwargs:
         """
-        self.model = model
+        self.request = request
         self.schema = schema
+        self.model = model
         self.url_id_field = url_id_field
         self.id_name_field = id_name_field
         self.disable_collection_count: bool = disable_collection_count
@@ -75,6 +56,38 @@ class BaseDataLayer:
     async def atomic_end(self, success: bool = True):
         raise NotImplementedError
 
+    def _unwrap_field_config(self, extra: Dict):
+        field_config_wrapper: Optional[TransferSaveWrapper] = extra.get("field_config")
+
+        if field_config_wrapper:
+            return field_config_wrapper.get_field_config()
+
+        return FieldConfig()
+
+    def _apply_client_generated_id(
+        self,
+        data_create: BaseJSONAPIItemInSchema,
+        model_kwargs: dict,
+    ):
+        """
+        :param data_create: the data validated by pydantic.
+        :param model_kwargs: the data validated by pydantic.
+        """
+        if data_create.id is None:
+            return model_kwargs
+
+        extra = data_create.__fields__["id"].field_info.extra
+        if extra.get("client_can_set_id"):
+            id_value = data_create.id
+            field_config = self._unwrap_field_config(extra)
+
+            if field_config.cast_type:
+                id_value = field_config.cast_type(id_value)
+
+            model_kwargs["id"] = id_value
+
+        return model_kwargs
+
     async def create_object(self, data_create: BaseJSONAPIItemInSchema, view_kwargs: dict) -> TypeModel:
         """
         Create an object
@@ -528,13 +541,3 @@ class BaseDataLayer:
         :param view_kwargs: kwargs from the resource view
         """
         raise NotImplementedError
-
-    def bound_rewritable_methods(self, methods):
-        """
-        Bound additional methods to current instance
-
-        :param class methods: methods
-        """
-        for key, value in methods.items():
-            if key in self.REWRITABLE_METHODS:
-                setattr(self, key, types.MethodType(value, self))
